#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include <imago2.h>
#include "cam.h"

void disp(void);
void render(float t, float pan_x, float pan_y);
void worm(float t, float z, float *tx, float *ty);
void cam_stereo_view_matrix(int eye);
void cam_stereo_proj_matrix(int eye);
unsigned int load_texture(const char *fname);
void reshape(int x, int y);
void keyb(unsigned char key, int x, int y);
void testimg(int left);

unsigned int tex;

int stereo, fullscr;
int show_testimg;

/* construction parameters */
int sides = 24;
int segm = 20;
float ring_height = 0.5;

int main(int argc, char **argv)
{
	int i;
	float fog_color[] = {0.6, 0.6, 0.6, 1};
	float lpos[] = {-1, 1, 2, 0};

	glutInitWindowSize(800, 600);
	glutInit(&argc, argv);

	for(i=1; i<argc; i++) {
		if(strcmp(argv[i], "-s") == 0) {
			stereo = 1;
		} else if(strcmp(argv[i], "-f") == 0) {
			fullscr = 1;
		}
	}

	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | (stereo ? GLUT_STEREO : 0));
	glutCreateWindow("stereo tunnel");

	if(fullscr) {
		glutFullScreen();
	}

	glutDisplayFunc(disp);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyb);
	glutIdleFunc(glutPostRedisplay);

	glEnable(GL_CULL_FACE);

	glEnable(GL_FOG);
	glFogf(GL_FOG_DENSITY, 0.2);
	glFogfv(GL_FOG_COLOR, fog_color);
	glClearColor(0.6, 0.6, 0.6, 1);

	if(!(tex = load_texture("tiles.jpg"))) {
		return 1;
	}

	glEnable(GL_LIGHT0);
	glLightfv(GL_LIGHT0, GL_POSITION, lpos);

	assert(glGetError() == GL_NO_ERROR);
	glutMainLoop();
	return 0;
}


void disp(void)
{
	unsigned long msec = glutGet(GLUT_ELAPSED_TIME);
	float sec = (float)msec / 1000.0;
	float pan_x, pan_y, z;

	z = ring_height * (segm - 1);
	worm(sec, z, &pan_x, &pan_y);

	if(stereo) {
		/* left eye */
		glDrawBuffer(GL_BACK_LEFT);

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		cam_stereo_proj_matrix(CAM_LEFT);

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		cam_stereo_view_matrix(CAM_LEFT);

		if(show_testimg) {
			testimg(0);
		} else {
			render(sec, pan_x, pan_y);
		}

		/* right eye */
		glDrawBuffer(GL_BACK_RIGHT);

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		cam_stereo_proj_matrix(CAM_RIGHT);

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		cam_stereo_view_matrix(CAM_RIGHT);

		if(show_testimg) {
			testimg(1);
		} else {
			render(sec, pan_x, pan_y);
		}
	} else {

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		cam_view_matrix();

		render(sec, pan_x, pan_y);
	}

	glutSwapBuffers();
	assert(glGetError() == GL_NO_ERROR);
}

void render(float t, float pan_x, float pan_y)
{
	int i, j;
	float tx, ty, prev_tx, prev_ty;

	glClear(GL_COLOR_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glTranslatef(-pan_x, -pan_y, -1.1 * ring_height * segm);

	glMatrixMode(GL_TEXTURE);
	glLoadIdentity();
	glTranslatef(0, -t * 1.5, 0);

	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, tex);

	glBegin(GL_QUAD_STRIP);
	glColor3f(1, 1, 1);

	prev_tx = prev_ty = 0.0;

	for(i=0; i<segm; i++) {
		float z = ring_height * (float)i;
		float v = (float)i / (float)segm;

		worm(t, z, &tx, &ty);

		for(j=0; j<sides; j++) {
			float u = (float)j / (float)(sides - 2);
			float theta = 2.0 * M_PI * u;

			float x = cos(theta);
			float y = sin(theta);

			glTexCoord2f(u * 2.0, v * 8.0);
			glVertex3f(x + prev_tx, y + prev_ty, z);
			glTexCoord2f(u * 2.0, (v + (1.0 / (float)segm)) * 8.0);
			glVertex3f(x + tx, y + ty, z + ring_height);
		}
		prev_tx = tx;
		prev_ty = ty;
	}
	glEnd();
	glDisable(GL_TEXTURE_2D);

	/*glEnable(GL_LIGHTING);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glTranslatef(0, 0, -1);

	glutSolidSphere(1.0, 12, 6);
	glPopMatrix();

	glDisable(GL_LIGHTING);*/
}

void worm(float t, float z, float *tx, float *ty)
{
	float x, y;
	x = sin(t) + cos(t + z) + sin(t * 2.0 + z) / 2.0;
	y = cos(t) + sin(t + z) + cos(t * 2.0 + z) / 2.0;

	*tx = x * 0.5;
	*ty = y * 0.5;
}

unsigned int load_texture(const char *fname)
{
	unsigned int tex;
	int xsz, ysz;
	void *pixels;

	if(!(pixels = img_load_pixels(fname, &xsz, &ysz, IMG_FMT_RGBA32))) {
		return 0;
	}

	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_2D, tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D, 0, 4, xsz, ysz, 0, GL_RGBA, GL_UNSIGNED_BYTE, pixels);
	free(pixels);

	return tex;
}

void reshape(int x, int y)
{
	glViewport(0, 0, x, y);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0, (float)x / (float)y, 1.0, 1000.0);
}

void keyb(unsigned char key, int x, int y)
{
	static int prev_xsz = -1, prev_ysz;

	switch(key) {
	case 27:
		exit(0);

	case 'f':
		if(prev_xsz != -1) {
			glutReshapeWindow(prev_xsz, prev_ysz);
			prev_xsz = -1;
		} else {
			prev_xsz = glutGet(GLUT_WINDOW_WIDTH);
			prev_ysz = glutGet(GLUT_WINDOW_HEIGHT);
			glutFullScreen();
		}
		break;

	case 't':
		show_testimg = !show_testimg;
		break;
	}
}

void testimg(int right)
{
	char *str = right ? "RIGHT ->" : "<- LEFT";

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glClear(GL_COLOR_BUFFER_BIT);

	glColor3f(0, 0, 0);
	glRasterPos2i(0, 0);
	while(*str) {
		glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, *str++);
	}
}
